/* --------------------------------- demo.c --------------------------------- */

/* This was part of the flight simulator 'fly8'.
 * Author: Eyal Lebedinsky (eyal@ise.canberra.edu.au).
*/

/* Demonstrate the use of a packet driver. This one is for ethernet type
 * packets. This one is for micro&soft C but should be reasonably portable.
 *
 * Packets are sent by calling packet_send(). Packets are recieved
 * asynchronously and collected on a chain (pack->head/tail) untill collected
 * by a call to pack_pick(). The application should periodically collect the
 * packets or a memory shortage will cause packets to be dropped.
 *
 * Note that the receiver() routine calls packet_new() which does a malloc();
 * this is potentially dangerous unless the malloc() is serialized. I usually
 * use something like my_malloc() which does a disable/enable around the real
 * call to malloc().
 *
 * Don't let the size of this demo confuse you, it is rather simple. All you
 * need to do is initialize the interface at the start and terminate it
 * when done. You may (just as well) use the interface to send packets
 * and you may also handle the ones received. If you do not wish to receive
 * any packets then you can either give the interface at the start a
 * very unusual packet_type of just return nulls from the (now very short)
 * receiver() routine.
 *
 * The setup allows you to have up to 4 'channels' to the packet driver.
 * Each  one can use a different network card or it may use the same one
 * but with a different packet_type. I use it successfuly with both an
 * NE2000 card and an ETHRSLIP over a serial line concurrently.
 *
 * The structures for PORT and PACKET are my own and you can design it to
 * fit your need. You don't have to keep all the driver's info in the
 * 'port' structure.
 *
 * To use this demo, first install your packet driver (I run
 * 'ne2000 0x66 5 0x300') then run 'demo'. Do the same on the other
 * machine. Now when you type a line on one machine it is sent to the
 * other and v.v. End the demo with a control-Z (DOS end-of-file character).
 *
 * This demo will NOT work with SLIP8250.COM but WILL work with
 * ETHERSL.COM which uses a proper ethernet interface.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <dos.h>
#include <conio.h>

#include "pktdrvr.h"

#define MY_OFF(p)	(((Ushort far *)&(p))[0])
#define MY_SEG(p)	(((Ushort far *)&(p))[1])

#define	TYPE_FLY8	0xf8f8

#define	PKSSIZE		1024		/* packet-driver's stack size */

#define	LADDRESS	6
#define PHEAD		(2*LADDRESS+2)

typedef unsigned char	Uchar;
typedef unsigned int	Uint;
typedef unsigned short	Ushort;
typedef unsigned long	Ulong;

typedef struct packet PACKET;
struct packet {
	PACKET	*next;
	short	length;		/* message length */
	Uchar	*data;
};

typedef struct port PORT;
struct port {
	int	flags;
#define POF_ON		0x0001
	void (interrupt far *pkint) (void);
	int	intno;			/* packet-driver interrupt */
	Uchar	address[LADDRESS];	/* my MAC address */
	int	handle;			/* packet-driver handle */
	char	packet_type[2];		/* eth_type for fly8 */
	PACKET	*pack;			/* packet being received */
	PACKET	*head;			/* packets received */
	PACKET	*tail;
	int	*stack;			/* packet-driver stack */
	int	version;
	int	class;
	int	type;
	int	number;
	char	*name;
	int	basic;
};

static PORT	ports[] = {
	{0, pkint0},
	{0, pkint1},
	{0, pkint2},
	{0, pkint3}
};
#define	NDEV	(sizeof(ports)/sizeof(ports[0])-1)

static int	nports = 0;		/* number of active ports */

static short far
disable (void)
{
	short	flags;

	_asm {
		pushf
		pop	flags
		cli
	}
	return (flags);
}

static void far
enable (short flags)
{
	_asm {
		push	flags
		popf
	}
}

static void far
packet_del (PACKET *pack)
{
	free (pack->data);
	free (pack);
}

static PACKET * far
packet_new (Ushort size)
{
	PACKET	*pack;

	if (size < PHEAD)
		return (0);

	if (0 != (pack = malloc (sizeof (PACKET)))) {
		pack->length = size;
		if (0 == (pack->data = malloc (size))) {
			free (pack);
			pack = 0;
		}
	}

	return (pack);
}

static void far
packet_deliver (PORT *port, PACKET *pack)
{
	pack->next = 0;
	if (port->tail)
		port->tail->next = pack;
	else
		port->head = pack;
	port->tail = pack;
}

static void far
receiver (int dev, Ushort di, Ushort si, Ushort bp, Ushort dx, Ushort cx,
	Ushort bx, Ushort ax, Ushort ds, Ushort es)
{
	PORT	*port;
	PACKET	*pack;
	char	*buff;

	switch (ax) {
	case 0:				/* Space allocate call */
		es = di = 0;
		if (dev < 0 || dev >= NDEV || cx > 256)
			break;
		port = &ports[dev];
		if (!(port->flags & POF_ON))
			break;
		if (port->pack) {		/* stray packet? */
			packet_del (port->pack);
			port->pack = 0;
		}
		if (0 == (pack = packet_new (cx)))
			break;
		port->pack = pack;
		buff = (char *)pack->data;
		es = MY_SEG (buff);
		di = MY_OFF (buff);
		break;
	case 1:				/* Packet complete call */
		if (dev < 0 || dev >= NDEV)
			break;
		port = &ports[dev];
		if (!(port->flags & POF_ON))
			break;
		if (0 == (pack = port->pack))
			break;

		packet_deliver (port, pack);
		port->pack = 0;
		break;
	default:
		break;
	}
}

static int far
find_driver (void)
{
	int	i;

	for (i = 0x0060; i < 0x0080; ++i)
		if (test_for_pd (i))
			return (i);
	return (-1);
}

static int far
com_init (int portno, int intno)
{
	int	rc;
	PORT	*port;
	char	msg[80], hex[10];

	if (portno < 0 || portno >= NDEV) {
		printf ("%d: bad port\n", portno);
		return (1);
	}
	port = &ports[portno];
	if (port->flags & POF_ON) {
		printf ("%d: already on\n", portno);
		return (1);
	}

	port->intno = intno;
	port->head = port->tail = port->pack = 0;

	if (-1 == port->intno)
		port->intno = find_driver ();
	else if (!test_for_pd (port->intno))
		port->intno = -1;
	if (-1 == port->intno) {
		printf ("%d: no driver\n", portno);
		return (1);
	}
	printf ("Intno 0x%x\n", port->intno);

	if (0 == (port->stack = (int *)calloc (PKSSIZE, sizeof (int)))) {
		printf ("%d: no mem\n", portno);
		return (1);
	}
	pkinit (portno, receiver, &port->stack[PKSSIZE]);

	port->packet_type[0] = (char)(0x0ff & (TYPE_FLY8>>8));
	port->packet_type[1] = (char)(0x0ff & (TYPE_FLY8   ));
	port->handle = access_type (port->intno, CL_ETHERNET, ANYTYPE, 0,
		(char far *)port->packet_type, 2, port->pkint);
	if (-1 == port->handle) {
		printf ("%d: no handle\n", portno);
		free (port->stack);
		port->stack = 0;
		return (1);
	}
	if (driver_info (port->intno, port->handle, &port->version,
			&port->class, &port->type, &port->number, &port->name,
			&port->basic)) {
		port->basic = 1;	/* what else ? */
	}
	printf ("Basic 0x%x\n", port->basic);

	port->flags |= POF_ON;
	++nports;

	rc = get_address (port->intno, port->handle, (char *)port->address,
						sizeof (port->address));
	strcpy (msg, "MAC ");
	if (rc) {
		strcat (msg, "failed ");
		sprintf (hex, "%0x", Derr);
		strcat (msg, hex);
		memset (port->address, 0, sizeof (port->address));
	} else {
		for (rc = 0; rc < LADDRESS; ++rc) {
			sprintf (hex, "%02x", port->address[rc]);
			strcat (msg, hex);
		}
	}
	printf ("%s\n", msg);

	return (0);
}

static void far
com_term (int portno)
{
	PORT	*port;

	if (portno < 0 || portno >= NDEV)
		return;
	port = &ports[portno];
	if (!(port->flags & POF_ON))
		return;
	release_type (port->intno, port->handle);
	if (port->pack) {
		packet_del (port->pack);
		port->pack = 0;
	}
	port->flags = 0;
	if (port->stack) {
		free (port->stack);
		port->stack = 0;
	}
}

static int far
packet_send (int portno, PACKET *pack)
{
	PORT	*port;

	if (0 == pack)
		return (0);

	if (portno < 0 || portno >= NDEV)
		return (1);
	port = &ports[portno];
	if (!(port->flags & POF_ON))
		return (1);

	memcpy (pack->data+LADDRESS, port->address, LADDRESS);	/* from */
	memcpy (pack->data+2*LADDRESS, port->packet_type, 2);	/* type */
	send_pkt (port->intno, (char *)pack->data, pack->length);

	return (0);
}

static PACKET * far
packet_pick (int portno)
{
	PORT	*port;
	PACKET	*p;
	short	flags;

	if (portno < 0 || portno >= NDEV)
		return (0);
	port = &ports[portno];
	if (!(port->flags & POF_ON))
		return (0);

	if (0 != (p = port->head)) {
		flags = disable ();
		port->head = p->next;
		if (0 == port->head)
			port->tail = 0;
		enable (flags);
	}
	return (p);
}

main ()
{
	PACKET	*pack;
	char	hex[3], msg[256];
	int	i;

	if (com_init (0, -1))
		exit (1);

	printf ("Enter 'quit' to stop.\n\n");
	fflush (stdout);

	for (;;) {
		if (kbhit ()) {
			if (0 == fgets (msg, sizeof (msg), stdin))
				break;
			if (!strcmp (msg, "quit\n"))		/* end */
				break;
			i = PHEAD + strlen (msg) + 1;
			if (0 == (pack = packet_new ((Ushort)i))) {
				printf ("no packet\n");
				break;
			}
			strcpy ((char *)pack->data+PHEAD, msg);
			memset (pack->data, 0xff, LADDRESS);	/* broadcast */
			packet_send (0, pack);
			packet_del (pack);
		}
		while (0 != (pack = packet_pick (0))) {
			msg[0] = '\0';
			for (i = 0; i < LADDRESS; ++i) {
				sprintf (hex, "%02x", pack->data[LADDRESS+i]);
				strcat (msg, hex);
			}
			printf ("msg from %s: %-.30s\n", msg, pack->data+PHEAD);
			packet_del (pack);
		}
	}

	com_term (0);

	exit (0);
	return (0);
}
